{-# OPTIONS_GHC -cpp -XMagicHash #-}
{- For Hugs, use the option -F"cpp -P -traditional" -}

module LinearScan.Resolve where


import Debug.Trace (trace, traceShow, traceShowId)
import qualified Prelude
import qualified Data.IntMap
import qualified Data.IntSet
import qualified Data.List
import qualified Data.Ord
import qualified Data.Functor.Identity
import qualified Hask.Utils

import qualified LinearScan.Allocate as Allocate
import qualified LinearScan.Applicative as Applicative
import qualified LinearScan.Blocks as Blocks
import qualified LinearScan.Functor as Functor
import qualified LinearScan.Graph as Graph
import qualified LinearScan.IntMap as IntMap
import qualified LinearScan.IntSet as IntSet
import qualified LinearScan.Interval as Interval
import qualified LinearScan.List1 as List1
import qualified LinearScan.LiveSets as LiveSets
import qualified LinearScan.Monad as Monad
import qualified LinearScan.Eqtype as Eqtype
import qualified LinearScan.Fintype as Fintype
import qualified LinearScan.Ssrbool as Ssrbool
import qualified LinearScan.Ssrnat as Ssrnat



#ifdef __GLASGOW_HASKELL__
import qualified GHC.Base as GHC.Base
import qualified GHC.Prim as GHC.Prim
#else
-- HUGS
import qualified LinearScan.IOExts as IOExts
#endif


#ifdef __GLASGOW_HASKELL__
--unsafeCoerce :: a -> b
unsafeCoerce = GHC.Base.unsafeCoerce#
#else
-- HUGS
--unsafeCoerce :: a -> b
unsafeCoerce = IOExts.unsafeCoerce
#endif

__ :: any
__ = Prelude.error "Logical or arity value used"

type PhysReg = Prelude.Int

data ResolvingMove =
   Move PhysReg Blocks.VarId PhysReg
 | Transfer PhysReg Blocks.VarId PhysReg
 | Spill PhysReg Blocks.VarId
 | Restore Blocks.VarId PhysReg
 | AllocReg Blocks.VarId PhysReg
 | FreeReg PhysReg Blocks.VarId
 | Looped ResolvingMove

coq_ResolvingMove_rect :: Prelude.Int -> (PhysReg -> Blocks.VarId -> PhysReg
                          -> a1) -> (PhysReg -> Blocks.VarId -> PhysReg ->
                          a1) -> (PhysReg -> Blocks.VarId -> a1) ->
                          (Blocks.VarId -> PhysReg -> a1) -> (Blocks.VarId ->
                          PhysReg -> a1) -> (PhysReg -> Blocks.VarId -> a1)
                          -> (ResolvingMove -> a1 -> a1) -> ResolvingMove ->
                          a1
coq_ResolvingMove_rect maxReg f f0 f1 f2 f3 f4 f5 r =
  case r of {
   Move p v p0 -> f p v p0;
   Transfer p v p0 -> f0 p v p0;
   Spill p v -> f1 p v;
   Restore v p -> f2 v p;
   AllocReg v p -> f3 v p;
   FreeReg p v -> f4 p v;
   Looped r0 -> f5 r0 (coq_ResolvingMove_rect maxReg f f0 f1 f2 f3 f4 f5 r0)}

data ResolvingMoveSet =
   RSMove Prelude.Int Blocks.VarId Prelude.Int
 | RSTransfer Prelude.Int Blocks.VarId Prelude.Int
 | RSSpill Prelude.Int Blocks.VarId
 | RSRestore Blocks.VarId Prelude.Int
 | RSAllocReg Blocks.VarId Prelude.Int
 | RSFreeReg Prelude.Int Blocks.VarId
 | RSAssignReg Blocks.VarId Prelude.Int
 | RSClearReg Prelude.Int Blocks.VarId
 | RSLooped ResolvingMoveSet

weakenResolvingMove :: Prelude.Int -> ResolvingMove -> ResolvingMoveSet
weakenResolvingMove maxReg x =
  case x of {
   Move fr fv tr -> RSMove ( fr) fv ( tr);
   Transfer fr fv tr -> RSTransfer ( fr) fv ( tr);
   Spill fr tv -> RSSpill ( fr) tv;
   Restore fv tr -> RSRestore fv ( tr);
   AllocReg fv tr -> RSAllocReg fv ( tr);
   FreeReg fr tv -> RSFreeReg ( fr) tv;
   Looped x0 -> RSLooped (weakenResolvingMove maxReg x0)}

eqResolvingMove :: Prelude.Int -> ResolvingMove -> ResolvingMove ->
                   Prelude.Bool
eqResolvingMove maxReg s1 s2 =
  case s1 of {
   Move fr1 fv1 tr1 ->
    case s2 of {
     Move fr2 fv2 tr2 ->
      (Prelude.&&)
        (Eqtype.eq_op (Fintype.ordinal_eqType maxReg) (unsafeCoerce fr1)
          (unsafeCoerce fr2))
        ((Prelude.&&)
          (Eqtype.eq_op Ssrnat.nat_eqType (unsafeCoerce fv1)
            (unsafeCoerce fv2))
          (Eqtype.eq_op (Fintype.ordinal_eqType maxReg) (unsafeCoerce tr1)
            (unsafeCoerce tr2)));
     _ -> Prelude.False};
   Transfer fr1 fv1 tr1 ->
    case s2 of {
     Transfer fr2 fv2 tr2 ->
      (Prelude.&&)
        (Eqtype.eq_op (Fintype.ordinal_eqType maxReg) (unsafeCoerce fr1)
          (unsafeCoerce fr2))
        ((Prelude.&&)
          (Eqtype.eq_op Ssrnat.nat_eqType (unsafeCoerce fv1)
            (unsafeCoerce fv2))
          (Eqtype.eq_op (Fintype.ordinal_eqType maxReg) (unsafeCoerce tr1)
            (unsafeCoerce tr2)));
     _ -> Prelude.False};
   Spill fr1 fv1 ->
    case s2 of {
     Spill fr2 fv2 ->
      (Prelude.&&)
        (Eqtype.eq_op (Fintype.ordinal_eqType maxReg) (unsafeCoerce fr1)
          (unsafeCoerce fr2))
        (Eqtype.eq_op Ssrnat.nat_eqType (unsafeCoerce fv1)
          (unsafeCoerce fv2));
     _ -> Prelude.False};
   Restore tv1 tr1 ->
    case s2 of {
     Restore tv2 tr2 ->
      (Prelude.&&)
        (Eqtype.eq_op Ssrnat.nat_eqType (unsafeCoerce tv1)
          (unsafeCoerce tv2))
        (Eqtype.eq_op (Fintype.ordinal_eqType maxReg) (unsafeCoerce tr1)
          (unsafeCoerce tr2));
     _ -> Prelude.False};
   AllocReg fv1 tr1 ->
    case s2 of {
     AllocReg fv2 tr2 ->
      (Prelude.&&)
        (Eqtype.eq_op Ssrnat.nat_eqType (unsafeCoerce fv1)
          (unsafeCoerce fv2))
        (Eqtype.eq_op (Fintype.ordinal_eqType maxReg) (unsafeCoerce tr1)
          (unsafeCoerce tr2));
     _ -> Prelude.False};
   FreeReg fr1 tv1 ->
    case s2 of {
     FreeReg fr2 tv2 ->
      (Prelude.&&)
        (Eqtype.eq_op (Fintype.ordinal_eqType maxReg) (unsafeCoerce fr1)
          (unsafeCoerce fr2))
        (Eqtype.eq_op Ssrnat.nat_eqType (unsafeCoerce tv1)
          (unsafeCoerce tv2));
     _ -> Prelude.False};
   Looped x ->
    case s2 of {
     Looped y -> eqResolvingMove maxReg x y;
     _ -> Prelude.False}}

eqResolvingMoveP :: Prelude.Int -> Eqtype.Equality__Coq_axiom ResolvingMove
eqResolvingMoveP maxReg _top_assumption_ =
  let {
   _evar_0_ = \fr1 fv1 tr1 _top_assumption_0 ->
    let {
     _evar_0_ = \fr2 fv2 tr2 ->
      let {
       _evar_0_ = \_ ->
        let {
         _evar_0_ = let {
                     _evar_0_ = \_ ->
                      let {
                       _evar_0_ = let {
                                   _evar_0_ = \_ ->
                                    let {_evar_0_ = Ssrbool.ReflectT} in
                                     _evar_0_}
                                  in
                                  let {_evar_0_0 = \_ -> Ssrbool.ReflectF} in
                                  case Eqtype.eqP
                                         (Fintype.ordinal_eqType maxReg) tr1
                                         tr2 of {
                                   Ssrbool.ReflectT -> _evar_0_ __;
                                   Ssrbool.ReflectF -> _evar_0_0 __}}
                      in
                       _evar_0_}
                    in
                    let {
                     _evar_0_0 = \_ ->
                      let {
                       _evar_0_0 = \_ ->
                        let {_evar_0_0 = Ssrbool.ReflectF} in  _evar_0_0}
                      in
                      let {_evar_0_1 = \_ -> Ssrbool.ReflectF} in
                      case Eqtype.eqP (Fintype.ordinal_eqType maxReg) tr1 tr2 of {
                       Ssrbool.ReflectT -> _evar_0_0 __;
                       Ssrbool.ReflectF -> _evar_0_1 __}}
                    in
                    case Eqtype.eqP Ssrnat.nat_eqType fv1 fv2 of {
                     Ssrbool.ReflectT -> _evar_0_ __;
                     Ssrbool.ReflectF -> _evar_0_0 __}}
        in
         _evar_0_}
      in
      let {
       _evar_0_0 = \_ ->
        let {
         _evar_0_0 = \_ ->
          let {
           _evar_0_0 = let {
                        _evar_0_0 = \_ ->
                         let {_evar_0_0 = Ssrbool.ReflectF} in  _evar_0_0}
                       in
                       let {_evar_0_1 = \_ -> Ssrbool.ReflectF} in
                       case Eqtype.eqP (Fintype.ordinal_eqType maxReg) tr1
                              tr2 of {
                        Ssrbool.ReflectT -> _evar_0_0 __;
                        Ssrbool.ReflectF -> _evar_0_1 __}}
          in
           _evar_0_0}
        in
        let {
         _evar_0_1 = \_ ->
          let {
           _evar_0_1 = \_ ->
            let {_evar_0_1 = Ssrbool.ReflectF} in  _evar_0_1}
          in
          let {_evar_0_2 = \_ -> Ssrbool.ReflectF} in
          case Eqtype.eqP (Fintype.ordinal_eqType maxReg) tr1 tr2 of {
           Ssrbool.ReflectT -> _evar_0_1 __;
           Ssrbool.ReflectF -> _evar_0_2 __}}
        in
        case Eqtype.eqP Ssrnat.nat_eqType fv1 fv2 of {
         Ssrbool.ReflectT -> _evar_0_0 __;
         Ssrbool.ReflectF -> _evar_0_1 __}}
      in
      case Eqtype.eqP (Fintype.ordinal_eqType maxReg) fr1 fr2 of {
       Ssrbool.ReflectT -> _evar_0_ __;
       Ssrbool.ReflectF -> _evar_0_0 __}}
    in
    let {_evar_0_0 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_1 = \fr2 fv2 -> Ssrbool.ReflectF} in
    let {_evar_0_2 = \tv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_3 = \fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_4 = \fr2 tv2 -> Ssrbool.ReflectF} in
    let {_evar_0_5 = \y -> Ssrbool.ReflectF} in
    case _top_assumption_0 of {
     Move x x0 x1 -> unsafeCoerce _evar_0_ x x0 x1;
     Transfer x x0 x1 -> _evar_0_0 x x0 x1;
     Spill x x0 -> _evar_0_1 x x0;
     Restore x x0 -> _evar_0_2 x x0;
     AllocReg x x0 -> _evar_0_3 x x0;
     FreeReg x x0 -> _evar_0_4 x x0;
     Looped x -> _evar_0_5 x}}
  in
  let {
   _evar_0_0 = \fr1 fv1 tr1 _top_assumption_0 ->
    let {_evar_0_0 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {
     _evar_0_1 = \fr2 fv2 tr2 ->
      let {
       _evar_0_1 = \_ ->
        let {
         _evar_0_1 = let {
                      _evar_0_1 = \_ ->
                       let {
                        _evar_0_1 = let {
                                     _evar_0_1 = \_ ->
                                      let {_evar_0_1 = Ssrbool.ReflectT} in
                                       _evar_0_1}
                                    in
                                    let {_evar_0_2 = \_ -> Ssrbool.ReflectF}
                                    in
                                    case Eqtype.eqP
                                           (Fintype.ordinal_eqType maxReg)
                                           tr1 tr2 of {
                                     Ssrbool.ReflectT -> _evar_0_1 __;
                                     Ssrbool.ReflectF -> _evar_0_2 __}}
                       in
                        _evar_0_1}
                     in
                     let {
                      _evar_0_2 = \_ ->
                       let {
                        _evar_0_2 = \_ ->
                         let {_evar_0_2 = Ssrbool.ReflectF} in  _evar_0_2}
                       in
                       let {_evar_0_3 = \_ -> Ssrbool.ReflectF} in
                       case Eqtype.eqP (Fintype.ordinal_eqType maxReg) tr1
                              tr2 of {
                        Ssrbool.ReflectT -> _evar_0_2 __;
                        Ssrbool.ReflectF -> _evar_0_3 __}}
                     in
                     case Eqtype.eqP Ssrnat.nat_eqType fv1 fv2 of {
                      Ssrbool.ReflectT -> _evar_0_1 __;
                      Ssrbool.ReflectF -> _evar_0_2 __}}
        in
         _evar_0_1}
      in
      let {
       _evar_0_2 = \_ ->
        let {
         _evar_0_2 = \_ ->
          let {
           _evar_0_2 = let {
                        _evar_0_2 = \_ ->
                         let {_evar_0_2 = Ssrbool.ReflectF} in  _evar_0_2}
                       in
                       let {_evar_0_3 = \_ -> Ssrbool.ReflectF} in
                       case Eqtype.eqP (Fintype.ordinal_eqType maxReg) tr1
                              tr2 of {
                        Ssrbool.ReflectT -> _evar_0_2 __;
                        Ssrbool.ReflectF -> _evar_0_3 __}}
          in
           _evar_0_2}
        in
        let {
         _evar_0_3 = \_ ->
          let {
           _evar_0_3 = \_ ->
            let {_evar_0_3 = Ssrbool.ReflectF} in  _evar_0_3}
          in
          let {_evar_0_4 = \_ -> Ssrbool.ReflectF} in
          case Eqtype.eqP (Fintype.ordinal_eqType maxReg) tr1 tr2 of {
           Ssrbool.ReflectT -> _evar_0_3 __;
           Ssrbool.ReflectF -> _evar_0_4 __}}
        in
        case Eqtype.eqP Ssrnat.nat_eqType fv1 fv2 of {
         Ssrbool.ReflectT -> _evar_0_2 __;
         Ssrbool.ReflectF -> _evar_0_3 __}}
      in
      case Eqtype.eqP (Fintype.ordinal_eqType maxReg) fr1 fr2 of {
       Ssrbool.ReflectT -> _evar_0_1 __;
       Ssrbool.ReflectF -> _evar_0_2 __}}
    in
    let {_evar_0_2 = \fr2 fv2 -> Ssrbool.ReflectF} in
    let {_evar_0_3 = \tv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_4 = \fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_5 = \fr2 tv2 -> Ssrbool.ReflectF} in
    let {_evar_0_6 = \y -> Ssrbool.ReflectF} in
    case _top_assumption_0 of {
     Move x x0 x1 -> _evar_0_0 x x0 x1;
     Transfer x x0 x1 -> unsafeCoerce _evar_0_1 x x0 x1;
     Spill x x0 -> _evar_0_2 x x0;
     Restore x x0 -> _evar_0_3 x x0;
     AllocReg x x0 -> _evar_0_4 x x0;
     FreeReg x x0 -> _evar_0_5 x x0;
     Looped x -> _evar_0_6 x}}
  in
  let {
   _evar_0_1 = \fr1 fv1 _top_assumption_0 ->
    let {_evar_0_1 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_2 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {
     _evar_0_3 = \fr2 fv2 ->
      let {
       _evar_0_3 = \_ ->
        let {
         _evar_0_3 = let {
                      _evar_0_3 = \_ ->
                       let {_evar_0_3 = Ssrbool.ReflectT} in  _evar_0_3}
                     in
                     let {_evar_0_4 = \_ -> Ssrbool.ReflectF} in
                     case Eqtype.eqP Ssrnat.nat_eqType fv1 fv2 of {
                      Ssrbool.ReflectT -> _evar_0_3 __;
                      Ssrbool.ReflectF -> _evar_0_4 __}}
        in
         _evar_0_3}
      in
      let {
       _evar_0_4 = \_ ->
        let {
         _evar_0_4 = \_ -> let {_evar_0_4 = Ssrbool.ReflectF} in  _evar_0_4}
        in
        let {_evar_0_5 = \_ -> Ssrbool.ReflectF} in
        case Eqtype.eqP Ssrnat.nat_eqType fv1 fv2 of {
         Ssrbool.ReflectT -> _evar_0_4 __;
         Ssrbool.ReflectF -> _evar_0_5 __}}
      in
      case Eqtype.eqP (Fintype.ordinal_eqType maxReg) fr1 fr2 of {
       Ssrbool.ReflectT -> _evar_0_3 __;
       Ssrbool.ReflectF -> _evar_0_4 __}}
    in
    let {_evar_0_4 = \tv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_5 = \fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_6 = \fr2 tv2 -> Ssrbool.ReflectF} in
    let {_evar_0_7 = \y -> Ssrbool.ReflectF} in
    case _top_assumption_0 of {
     Move x x0 x1 -> _evar_0_1 x x0 x1;
     Transfer x x0 x1 -> _evar_0_2 x x0 x1;
     Spill x x0 -> unsafeCoerce _evar_0_3 x x0;
     Restore x x0 -> _evar_0_4 x x0;
     AllocReg x x0 -> _evar_0_5 x x0;
     FreeReg x x0 -> _evar_0_6 x x0;
     Looped x -> _evar_0_7 x}}
  in
  let {
   _evar_0_2 = \tv1 tr1 _top_assumption_0 ->
    let {_evar_0_2 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_3 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_4 = \fr2 fv2 -> Ssrbool.ReflectF} in
    let {
     _evar_0_5 = \tv2 tr2 ->
      let {
       _evar_0_5 = \_ ->
        let {
         _evar_0_5 = let {
                      _evar_0_5 = \_ ->
                       let {_evar_0_5 = Ssrbool.ReflectT} in  _evar_0_5}
                     in
                     let {_evar_0_6 = \_ -> Ssrbool.ReflectF} in
                     case Eqtype.eqP (Fintype.ordinal_eqType maxReg) tr1 tr2 of {
                      Ssrbool.ReflectT -> _evar_0_5 __;
                      Ssrbool.ReflectF -> _evar_0_6 __}}
        in
         _evar_0_5}
      in
      let {
       _evar_0_6 = \_ ->
        let {
         _evar_0_6 = \_ -> let {_evar_0_6 = Ssrbool.ReflectF} in  _evar_0_6}
        in
        let {_evar_0_7 = \_ -> Ssrbool.ReflectF} in
        case Eqtype.eqP (Fintype.ordinal_eqType maxReg) tr1 tr2 of {
         Ssrbool.ReflectT -> _evar_0_6 __;
         Ssrbool.ReflectF -> _evar_0_7 __}}
      in
      case Eqtype.eqP Ssrnat.nat_eqType tv1 tv2 of {
       Ssrbool.ReflectT -> _evar_0_5 __;
       Ssrbool.ReflectF -> _evar_0_6 __}}
    in
    let {_evar_0_6 = \fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_7 = \fr2 tv2 -> Ssrbool.ReflectF} in
    let {_evar_0_8 = \y -> Ssrbool.ReflectF} in
    case _top_assumption_0 of {
     Move x x0 x1 -> _evar_0_2 x x0 x1;
     Transfer x x0 x1 -> _evar_0_3 x x0 x1;
     Spill x x0 -> _evar_0_4 x x0;
     Restore x x0 -> unsafeCoerce _evar_0_5 x x0;
     AllocReg x x0 -> _evar_0_6 x x0;
     FreeReg x x0 -> _evar_0_7 x x0;
     Looped x -> _evar_0_8 x}}
  in
  let {
   _evar_0_3 = \fv1 tr1 _top_assumption_0 ->
    let {_evar_0_3 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_4 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_5 = \fr2 fv2 -> Ssrbool.ReflectF} in
    let {_evar_0_6 = \tv2 tr2 -> Ssrbool.ReflectF} in
    let {
     _evar_0_7 = \fv2 tr2 ->
      let {
       _evar_0_7 = \_ ->
        let {
         _evar_0_7 = let {
                      _evar_0_7 = \_ ->
                       let {_evar_0_7 = Ssrbool.ReflectT} in  _evar_0_7}
                     in
                     let {_evar_0_8 = \_ -> Ssrbool.ReflectF} in
                     case Eqtype.eqP (Fintype.ordinal_eqType maxReg) tr1 tr2 of {
                      Ssrbool.ReflectT -> _evar_0_7 __;
                      Ssrbool.ReflectF -> _evar_0_8 __}}
        in
         _evar_0_7}
      in
      let {
       _evar_0_8 = \_ ->
        let {
         _evar_0_8 = \_ -> let {_evar_0_8 = Ssrbool.ReflectF} in  _evar_0_8}
        in
        let {_evar_0_9 = \_ -> Ssrbool.ReflectF} in
        case Eqtype.eqP (Fintype.ordinal_eqType maxReg) tr1 tr2 of {
         Ssrbool.ReflectT -> _evar_0_8 __;
         Ssrbool.ReflectF -> _evar_0_9 __}}
      in
      case Eqtype.eqP Ssrnat.nat_eqType fv1 fv2 of {
       Ssrbool.ReflectT -> _evar_0_7 __;
       Ssrbool.ReflectF -> _evar_0_8 __}}
    in
    let {_evar_0_8 = \fr2 tv2 -> Ssrbool.ReflectF} in
    let {_evar_0_9 = \y -> Ssrbool.ReflectF} in
    case _top_assumption_0 of {
     Move x x0 x1 -> _evar_0_3 x x0 x1;
     Transfer x x0 x1 -> _evar_0_4 x x0 x1;
     Spill x x0 -> _evar_0_5 x x0;
     Restore x x0 -> _evar_0_6 x x0;
     AllocReg x x0 -> unsafeCoerce _evar_0_7 x x0;
     FreeReg x x0 -> _evar_0_8 x x0;
     Looped x -> _evar_0_9 x}}
  in
  let {
   _evar_0_4 = \fr1 tv1 _top_assumption_0 ->
    let {_evar_0_4 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_5 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_6 = \fr2 fv2 -> Ssrbool.ReflectF} in
    let {_evar_0_7 = \tv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_8 = \fv2 tr2 -> Ssrbool.ReflectF} in
    let {
     _evar_0_9 = \fr2 tv2 ->
      let {
       _evar_0_9 = \_ ->
        let {
         _evar_0_9 = let {
                      _evar_0_9 = \_ ->
                       let {_evar_0_9 = Ssrbool.ReflectT} in  _evar_0_9}
                     in
                     let {_evar_0_10 = \_ -> Ssrbool.ReflectF} in
                     case Eqtype.eqP Ssrnat.nat_eqType tv1 tv2 of {
                      Ssrbool.ReflectT -> _evar_0_9 __;
                      Ssrbool.ReflectF -> _evar_0_10 __}}
        in
         _evar_0_9}
      in
      let {
       _evar_0_10 = \_ ->
        let {
         _evar_0_10 = \_ ->
          let {_evar_0_10 = Ssrbool.ReflectF} in  _evar_0_10}
        in
        let {_evar_0_11 = \_ -> Ssrbool.ReflectF} in
        case Eqtype.eqP Ssrnat.nat_eqType tv1 tv2 of {
         Ssrbool.ReflectT -> _evar_0_10 __;
         Ssrbool.ReflectF -> _evar_0_11 __}}
      in
      case Eqtype.eqP (Fintype.ordinal_eqType maxReg) fr1 fr2 of {
       Ssrbool.ReflectT -> _evar_0_9 __;
       Ssrbool.ReflectF -> _evar_0_10 __}}
    in
    let {_evar_0_10 = \y -> Ssrbool.ReflectF} in
    case _top_assumption_0 of {
     Move x x0 x1 -> _evar_0_4 x x0 x1;
     Transfer x x0 x1 -> _evar_0_5 x x0 x1;
     Spill x x0 -> _evar_0_6 x x0;
     Restore x x0 -> _evar_0_7 x x0;
     AllocReg x x0 -> _evar_0_8 x x0;
     FreeReg x x0 -> unsafeCoerce _evar_0_9 x x0;
     Looped x -> _evar_0_10 x}}
  in
  let {
   _evar_0_5 = \x iHx _top_assumption_0 ->
    let {_evar_0_5 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_6 = \fr2 fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_7 = \fr2 fv2 -> Ssrbool.ReflectF} in
    let {_evar_0_8 = \tv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_9 = \fv2 tr2 -> Ssrbool.ReflectF} in
    let {_evar_0_10 = \fr2 tv2 -> Ssrbool.ReflectF} in
    let {
     _evar_0_11 = \y ->
      let {iHx0 = iHx y} in
      let {_evar_0_11 = \_ -> Ssrbool.ReflectT} in
      let {_evar_0_12 = \_ -> Ssrbool.ReflectF} in
      case iHx0 of {
       Ssrbool.ReflectT -> _evar_0_11 __;
       Ssrbool.ReflectF -> _evar_0_12 __}}
    in
    case _top_assumption_0 of {
     Move x0 x1 x2 -> _evar_0_5 x0 x1 x2;
     Transfer x0 x1 x2 -> _evar_0_6 x0 x1 x2;
     Spill x0 x1 -> _evar_0_7 x0 x1;
     Restore x0 x1 -> _evar_0_8 x0 x1;
     AllocReg x0 x1 -> _evar_0_9 x0 x1;
     FreeReg x0 x1 -> _evar_0_10 x0 x1;
     Looped x0 -> _evar_0_11 x0}}
  in
  coq_ResolvingMove_rect maxReg (unsafeCoerce _evar_0_)
    (unsafeCoerce _evar_0_0) (unsafeCoerce _evar_0_1)
    (unsafeCoerce _evar_0_2) (unsafeCoerce _evar_0_3)
    (unsafeCoerce _evar_0_4) _evar_0_5 _top_assumption_

coq_ResolvingMove_eqMixin :: Prelude.Int -> Eqtype.Equality__Coq_mixin_of
                             ResolvingMove
coq_ResolvingMove_eqMixin maxReg =
  Eqtype.Equality__Mixin (eqResolvingMove maxReg) (eqResolvingMoveP maxReg)

coq_ResolvingMove_eqType :: Prelude.Int -> Eqtype.Equality__Coq_type
coq_ResolvingMove_eqType maxReg =
  unsafeCoerce (coq_ResolvingMove_eqMixin maxReg)

data ResGraphNode =
   RegNode PhysReg
 | VarNode Blocks.VarId
 | VirtNode ResGraphNode

coq_ResGraphNode_rect :: Prelude.Int -> (PhysReg -> a1) -> (Blocks.VarId ->
                         a1) -> (ResGraphNode -> a1 -> a1) -> ResGraphNode ->
                         a1
coq_ResGraphNode_rect maxReg f f0 f1 r =
  case r of {
   RegNode p -> f p;
   VarNode v -> f0 v;
   VirtNode r0 -> f1 r0 (coq_ResGraphNode_rect maxReg f f0 f1 r0)}

eqResGraphNode :: Prelude.Int -> ResGraphNode -> ResGraphNode -> Prelude.Bool
eqResGraphNode maxReg s1 s2 =
  case s1 of {
   RegNode r1 ->
    case s2 of {
     RegNode r2 ->
      Eqtype.eq_op (Fintype.ordinal_eqType maxReg) (unsafeCoerce r1)
        (unsafeCoerce r2);
     _ -> Prelude.False};
   VarNode v1 ->
    case s2 of {
     VarNode v2 ->
      Eqtype.eq_op Ssrnat.nat_eqType (unsafeCoerce v1) (unsafeCoerce v2);
     _ -> Prelude.False};
   VirtNode n1 ->
    case s2 of {
     VirtNode n2 -> eqResGraphNode maxReg n1 n2;
     _ -> Prelude.False}}

eqResGraphNodeP :: Prelude.Int -> Eqtype.Equality__Coq_axiom ResGraphNode
eqResGraphNodeP maxReg _top_assumption_ =
  let {
   _evar_0_ = \r1 _top_assumption_0 ->
    let {
     _evar_0_ = \r2 ->
      let {_evar_0_ = \_ -> let {_evar_0_ = Ssrbool.ReflectT} in  _evar_0_}
      in
      let {_evar_0_0 = \_ -> Ssrbool.ReflectF} in
      case Eqtype.eqP (Fintype.ordinal_eqType maxReg) r1 r2 of {
       Ssrbool.ReflectT -> _evar_0_ __;
       Ssrbool.ReflectF -> _evar_0_0 __}}
    in
    let {_evar_0_0 = \v2 -> Ssrbool.ReflectF} in
    let {_evar_0_1 = \n2 -> Ssrbool.ReflectF} in
    case _top_assumption_0 of {
     RegNode x -> unsafeCoerce _evar_0_ x;
     VarNode x -> _evar_0_0 x;
     VirtNode x -> _evar_0_1 x}}
  in
  let {
   _evar_0_0 = \v1 _top_assumption_0 ->
    let {_evar_0_0 = \r2 -> Ssrbool.ReflectF} in
    let {
     _evar_0_1 = \v2 ->
      let {
       _evar_0_1 = \_ -> let {_evar_0_1 = Ssrbool.ReflectT} in  _evar_0_1}
      in
      let {_evar_0_2 = \_ -> Ssrbool.ReflectF} in
      case Eqtype.eqP Ssrnat.nat_eqType v1 v2 of {
       Ssrbool.ReflectT -> _evar_0_1 __;
       Ssrbool.ReflectF -> _evar_0_2 __}}
    in
    let {_evar_0_2 = \n2 -> Ssrbool.ReflectF} in
    case _top_assumption_0 of {
     RegNode x -> _evar_0_0 x;
     VarNode x -> unsafeCoerce _evar_0_1 x;
     VirtNode x -> _evar_0_2 x}}
  in
  let {
   _evar_0_1 = \n1 iHn _top_assumption_0 ->
    let {_evar_0_1 = \r2 -> Ssrbool.ReflectF} in
    let {_evar_0_2 = \v2 -> Ssrbool.ReflectF} in
    let {
     _evar_0_3 = \n2 ->
      let {iHn0 = iHn n2} in
      let {_evar_0_3 = \_ -> Ssrbool.ReflectT} in
      let {_evar_0_4 = \_ -> Ssrbool.ReflectF} in
      case iHn0 of {
       Ssrbool.ReflectT -> _evar_0_3 __;
       Ssrbool.ReflectF -> _evar_0_4 __}}
    in
    case _top_assumption_0 of {
     RegNode x -> _evar_0_1 x;
     VarNode x -> _evar_0_2 x;
     VirtNode x -> _evar_0_3 x}}
  in
  coq_ResGraphNode_rect maxReg (unsafeCoerce _evar_0_)
    (unsafeCoerce _evar_0_0) _evar_0_1 _top_assumption_

coq_ResGraphNode_eqMixin :: Prelude.Int -> Eqtype.Equality__Coq_mixin_of
                            ResGraphNode
coq_ResGraphNode_eqMixin maxReg =
  Eqtype.Equality__Mixin (eqResGraphNode maxReg) (eqResGraphNodeP maxReg)

coq_ResGraphNode_eqType :: Prelude.Int -> Eqtype.Equality__Coq_type
coq_ResGraphNode_eqType maxReg =
  unsafeCoerce (coq_ResGraphNode_eqMixin maxReg)

determineNodes :: Prelude.Int -> ResolvingMove -> (,) ResGraphNode
                  ResGraphNode
determineNodes maxReg x =
  case x of {
   Move fr fv tr -> (,) (RegNode tr) (RegNode fr);
   Transfer fr fv tr -> (,) (RegNode tr) (RegNode fr);
   Spill fr tv -> (,) (VirtNode (VarNode tv)) (RegNode fr);
   Restore fv tr -> (,) (RegNode tr) (VarNode fv);
   AllocReg fv tr -> (,) (RegNode tr) (VarNode fv);
   FreeReg fr tv -> (,) (VirtNode (VarNode tv)) (RegNode fr);
   Looped x0 -> determineNodes maxReg x0}

isMoveSplittable :: Prelude.Int -> ResolvingMove -> Prelude.Bool
isMoveSplittable maxReg x =
  case x of {
   Move p v p0 -> Prelude.True;
   Transfer p v p0 -> Prelude.True;
   _ -> Prelude.False}

splitMove :: Prelude.Int -> ResolvingMove -> [] ResolvingMove
splitMove maxReg x =
  case x of {
   Move fr fv tr -> (:) (Spill fr fv) ((:) (Restore fv tr) []);
   Transfer fr fv tr -> (:) (FreeReg fr fv) ((:) (AllocReg fv tr) []);
   Looped r -> (:) x [];
   _ -> (:) (Looped x) []}

resolvingMoves :: Prelude.Int -> ([] Allocate.Allocation) -> (Prelude.Maybe
                  IntSet.IntSet) -> Prelude.Int -> Prelude.Int ->
                  IntMap.IntMap ResolvingMove
resolvingMoves maxReg allocs liveIn from to =
  let {
   allocsAt = \pos ->
    Prelude.foldr (\a rest ->
      let {i = Allocate.intVal maxReg a} in
      case (Prelude.&&) ((Prelude.<=) (Interval.ibeg i) pos)
             ((Prelude.<=) ((Prelude.succ) pos) (Interval.iend i)) of {
       Prelude.True -> IntMap.coq_IntMap_insert (Interval.ivar i) a rest;
       Prelude.False -> rest}) IntMap.emptyIntMap allocs}
  in
  let {
   varNotLive = \vid ->
    case liveIn of {
     Prelude.Just ins -> Prelude.not (IntSet.coq_IntSet_member vid ins);
     Prelude.Nothing -> Prelude.False}}
  in
  IntMap.coq_IntMap_combine (\vid mx my ->
    case mx of {
     Prelude.Just x ->
      case my of {
       Prelude.Just y ->
        case Eqtype.eq_op
               (Eqtype.option_eqType (Fintype.ordinal_eqType maxReg))
               (unsafeCoerce (Allocate.intReg maxReg x))
               (unsafeCoerce (Allocate.intReg maxReg y)) of {
         Prelude.True -> Prelude.Nothing;
         Prelude.False ->
          case Allocate.intReg maxReg x of {
           Prelude.Just xr ->
            case Allocate.intReg maxReg y of {
             Prelude.Just yr -> Prelude.Just
              (case (Prelude.||) (varNotLive vid)
                      (Prelude.not (Ssrnat.odd to)) of {
                Prelude.True -> Transfer xr vid yr;
                Prelude.False -> Move xr vid yr});
             Prelude.Nothing -> Prelude.Just (Spill xr vid)};
           Prelude.Nothing ->
            case Allocate.intReg maxReg y of {
             Prelude.Just yr -> Prelude.Just
              (case (Prelude.||) (varNotLive vid)
                      (Prelude.not (Ssrnat.odd to)) of {
                Prelude.True -> AllocReg vid yr;
                Prelude.False -> Restore vid yr});
             Prelude.Nothing -> Prelude.Nothing}}};
       Prelude.Nothing ->
        case Allocate.intReg maxReg x of {
         Prelude.Just r -> Prelude.Just (FreeReg r vid);
         Prelude.Nothing -> Prelude.Nothing}};
     Prelude.Nothing ->
      case my of {
       Prelude.Just y ->
        case Allocate.intReg maxReg y of {
         Prelude.Just r -> Prelude.Just (AllocReg vid r);
         Prelude.Nothing -> Prelude.Nothing};
       Prelude.Nothing -> Prelude.Nothing}}) (allocsAt from) (allocsAt to)

determineMoves :: Prelude.Int -> ([] Allocate.Allocation) -> (Prelude.Maybe
                  IntSet.IntSet) -> Prelude.Int -> Prelude.Int -> []
                  ResolvingMove
determineMoves maxReg allocs liveIn from to =
  let {
   sortMoves = \x ->
    Prelude.map (\i -> Prelude.snd i)
      (Prelude.snd
        (Graph.topsort (coq_ResGraphNode_eqType maxReg)
          (coq_ResolvingMove_eqType maxReg) x
          (unsafeCoerce (isMoveSplittable maxReg))
          (unsafeCoerce (splitMove maxReg))))}
  in
  unsafeCoerce sortMoves
    (IntMap.coq_IntMap_foldr
      (unsafeCoerce
        (Graph.addEdge (coq_ResGraphNode_eqType maxReg)
          (coq_ResolvingMove_eqType maxReg)))
      (Graph.emptyGraph (coq_ResGraphNode_eqType maxReg)
        (coq_ResolvingMove_eqType maxReg)
        (unsafeCoerce (determineNodes maxReg)))
      (resolvingMoves maxReg allocs liveIn from to))

type BlockMoves = (,) ([] ResolvingMove) ([] ResolvingMove)

checkBlockBoundary :: Prelude.Int -> ([] Allocate.Allocation) ->
                      IntSet.IntSet -> Prelude.Int -> Prelude.Bool ->
                      (Prelude.Maybe LiveSets.BlockLiveSets) ->
                      LiveSets.BlockLiveSets -> (IntMap.IntMap BlockMoves) ->
                      IntMap.IntMap BlockMoves
checkBlockBoundary maxReg allocs liveIn bid in_from mfrom to mappings =
  let {
   moves = determineMoves maxReg allocs (Prelude.Just liveIn)
             (case mfrom of {
               Prelude.Just from -> LiveSets.blockLastOpId from;
               Prelude.Nothing -> 0}) (LiveSets.blockFirstOpId to)}
  in
  IntMap.coq_IntMap_alter (\mx ->
    case case mx of {
          Prelude.Just x -> x;
          Prelude.Nothing -> (,) [] []} of {
     (,) begs ends -> Prelude.Just
      (case in_from of {
        Prelude.True -> (,) begs moves;
        Prelude.False -> (,) moves ends})}) bid mappings

resolveDataFlow :: Prelude.Int -> (Monad.Monad a5) -> (Blocks.BlockInfo 
                   a5 a1 a2 a3 a4) -> ([] Allocate.Allocation) -> ([] 
                   a1) -> (IntMap.IntMap LiveSets.BlockLiveSets) -> a5
resolveDataFlow maxReg mDict binfo allocs blocks liveSets =
  Functor.fmap (Applicative.is_functor (Monad.is_applicative mDict))
    Prelude.fst
    (Monad.forFoldM mDict ((,) IntMap.emptyIntMap Prelude.True) blocks
      (\z b ->
      case z of {
       (,) mappings isFirst ->
        Monad.bind mDict (\bid ->
          case IntMap.coq_IntMap_lookup bid liveSets of {
           Prelude.Just from ->
            let {
             mappings' = case isFirst of {
                          Prelude.True ->
                           checkBlockBoundary maxReg allocs
                             (LiveSets.blockLiveIn from) bid Prelude.False
                             Prelude.Nothing from mappings;
                          Prelude.False -> mappings}}
            in
            Monad.bind mDict (\suxs ->
              let {
               in_from = (Prelude.<=) (Data.List.length suxs) ((Prelude.succ)
                           0)}
              in
              let {
               mappings'' = case Eqtype.eq_op Ssrnat.nat_eqType
                                   (unsafeCoerce (Data.List.length suxs))
                                   (unsafeCoerce 0) of {
                             Prelude.True -> mappings';
                             Prelude.False ->
                              List1.forFold mappings' suxs (\ms s_bid ->
                                case IntMap.coq_IntMap_lookup s_bid liveSets of {
                                 Prelude.Just to ->
                                  let {
                                   key = case in_from of {
                                          Prelude.True -> bid;
                                          Prelude.False -> s_bid}}
                                  in
                                  checkBlockBoundary maxReg allocs
                                    (LiveSets.blockLiveIn to) key in_from
                                    (Prelude.Just from) to ms;
                                 Prelude.Nothing -> ms})}}
              in
              Applicative.pure (Monad.is_applicative mDict) ((,) mappings''
                Prelude.False)) (Blocks.blockSuccessors mDict binfo b);
           Prelude.Nothing ->
            Applicative.pure (Monad.is_applicative mDict) ((,) mappings
              Prelude.False)}) (Blocks.blockId mDict binfo b)}))

